local LCCC = LibCodesCommonCode
local LMAA = LibMultiAccountAchievements
local LEJ = LibExtendedJournal
local RCR = Raidificator


--------------------------------------------------------------------------------
-- Extended Journal
--------------------------------------------------------------------------------

local TAB_NAME = "RaidificatorAchievements"
local FRAME = RaidificatorAchievementsFrame
local DATA_TYPE = 1
local SORT_TYPE = 1

local Initialized = false
local Dirtiness = 0
local ContextMenuItems = { }

function RCR.InitializeAchievements( )
	LEJ.RegisterTab(TAB_NAME, {
		title = SI_RCR_TITLE,
		subtitle = SI_RCR_SUBTITLE_ACHIEVEMENTS,
		order = 500,
		iconPrefix = "/esoui/art/journal/journal_tabicon_achievements_",
		control = FRAME,
		settingsPanel = RCR.settingsPanel,
		binding = "RCR_ACHIEVEMENTS",
		callbackShow = function( )
			RCR.LazyInitializeAchievements()
			RCR.RefreshAchievements(true)
		end,
	})
end

function RCR.LazyInitializeAchievements( )
	if (not Initialized) then
		Initialized = true

		RCR.achList = RaidificatorAchievementsList:New(FRAME, ContextMenuItems)

		local refresh = function( )
			Dirtiness = 1
			RCR.RefreshAchievements()
		end

		EVENT_MANAGER:RegisterForEvent(RCR.name, EVENT_ACHIEVEMENTS_UPDATED, refresh)
		EVENT_MANAGER:RegisterForEvent(RCR.name, EVENT_ACHIEVEMENT_AWARDED, refresh)
		EVENT_MANAGER:RegisterForEvent(RCR.name, EVENT_ACHIEVEMENT_UPDATED, refresh)
	end
end

function RCR.RefreshAchievements( noActiveCheck )
	if (Initialized and Dirtiness > 0 and (noActiveCheck or LEJ.IsTabActive(TAB_NAME))) then
		RCR.achList:RefreshAchievementState()
		Dirtiness = 0
	end
end


--------------------------------------------------------------------------------
-- API Acesss (overridden later if LMAA is present)
--------------------------------------------------------------------------------

local GetAchTimestamp = GetAchievementTimestamp
local IsAchComplete = IsAchievementComplete
local GetAchCriterion = GetAchievementCriterion
local GetAchLink = GetAchievementLink


--------------------------------------------------------------------------------
-- Register Context Menu
--------------------------------------------------------------------------------

function RCR.RegisterAchievementContextMenuItem( func )
	table.insert(ContextMenuItems, func)
end

RCR.RegisterAchievementContextMenuItem(function( data )
	return SI_RCR_LINK_INCOMPLETE, function( )
		local results = { }
		local remaining = 346
		local skipped = 0

		for _, ach in ipairs(data.achs) do
			if (not IsAchComplete(ach.id)) then
				local link = GetAchLink(ach.id, LINK_STYLE_BRACKETS)
				local newRemaining = remaining - zo_strlen(link)
				if (newRemaining >= 0) then
					table.insert(results, link)
					remaining = newRemaining
				else
					skipped = skipped + 1
				end
			end
		end

		if (#results > 0) then
			local message = table.concat(results, "")
			if (skipped > 0) then
				message = string.format("%s +%d", message, skipped)
			end
			ZO_LinkHandler_InsertLink(message)
		end
	end
end)

RCR.RegisterAchievementContextMenuItem(function( data )
	if (data.contentId) then
		return SI_RCR_SUBTITLE_HISTORY, function( )
			RCR.NewClearAdded(data.contentId)
			LEJ.Show("Raidificator")
		end
	end
end)


--------------------------------------------------------------------------------
-- Colors
--------------------------------------------------------------------------------

local COLOR_COMPLETE_BROWSER = 0x00FF00FF
local COLOR_COMPLETE_POPUP = 0x66CCFFFF
local COLOR_INCOMPLETE = 0x666666FF


--------------------------------------------------------------------------------
-- Achievement Data
--------------------------------------------------------------------------------

RCR.ACHIEVEMENT_FIELD_SIZE = 3
RCR.ACHIEVEMENT_FLAGS = {
	V = 0x01, -- Veteran
	C = 0x02, -- Clear
	H = 0x04, -- Hard Mode
	S = 0x08, -- Speed Run
	N = 0x10, -- No Death
	T = 0x20, -- Trifecta
}

local CATEGORY_DUNGEONS = 1
local CATEGORY_TRIALS = 2
local CATEGORY_ARENAS = 3

do
	local Data

	local function CheckCategory( zoneId, code, category, offset )
		local zone = RCR.ZONES[code][zoneId]
		if (zone) then
			return {
				category = category,
				year = zone[offset],
				trophy = zone[offset + 1]
			}
		end
	end

	function RCR.GetAchievementData( )
		if (not Data) then
			Data = { }

			for _, encoded in ipairs({ zo_strsplit(",", LCCC.Unchunk(RCR.ACHIEVEMENT_DATA)) }) do
				local length = zo_strlen(encoded)
				local zoneId, i = LCCC.ReadAndDecode(encoded, 1, RCR.ACHIEVEMENT_FIELD_SIZE)
				local zoneName = LCCC.GetZoneName(zoneId)

				if (zoneName ~= "") then
					local entry = CheckCategory(zoneId, "D", CATEGORY_DUNGEONS, 3) or CheckCategory(zoneId, "T", CATEGORY_TRIALS, 1) or CheckCategory(zoneId, "A", CATEGORY_ARENAS, 1)

					if (entry) then
						entry.zoneId = zoneId
						entry.name = zoneName
						entry.achs = { }

						while (i < length) do
							local achievementId, flags
							achievementId, i = LCCC.ReadAndDecode(encoded, i, RCR.ACHIEVEMENT_FIELD_SIZE)
							flags, i = LCCC.ReadAndDecode(encoded, i, 1)
							if (GetAchievementNumCriteria(achievementId) > 0) then
								table.insert(entry.achs, { id = achievementId, flags = flags })
							end
						end

						if (entry.category ~= CATEGORY_DUNGEONS) then
							for _, content in ipairs(RCR.SCORED_CONTENT) do
								if (entry.zoneId == content[2]) then
									entry.contentId = content[1]
								end
							end
						end

						table.insert(Data, entry)
					end
				end
			end
		end

		return Data
	end
end


--------------------------------------------------------------------------------
-- Helpers
--------------------------------------------------------------------------------

local function CheckFlag( value, flag )
	flag = RCR.ACHIEVEMENT_FLAGS[flag]
	return BitAnd(value, flag) == flag
end

local function AddToTally( tallies, flags )
	for k in pairs(tallies) do
		if (RCR.ACHIEVEMENT_FLAGS[k] and CheckFlag(flags, k)) then
			tallies[k] = tallies[k] + 1
		end
	end
end

local function SetRatioColor( control, ratio )
	control.nonRecolorable = true
	if (ratio == 1) then
		control:SetColor(LCCC.Int32ToRGBA(COLOR_COMPLETE_BROWSER))
	elseif (ratio == 0) then
		control:SetColor(LCCC.Int32ToRGBA(COLOR_INCOMPLETE))
	elseif (ratio < 0) then
		control:SetColor(0, 0, 0, 0)
	else
		control:SetColor(LCCC.HSLToRGB((1 / 3 + ratio) / 6, 1, 0.5, 1))
	end
end


--------------------------------------------------------------------------------
-- RaidificatorAchievementsList
--------------------------------------------------------------------------------

RaidificatorAchievementsList = ExtendedJournalSortFilterList:Subclass()
local RaidificatorAchievementsList = RaidificatorAchievementsList

function RaidificatorAchievementsList:Setup( )
	ZO_ScrollList_AddDataType(self.list, DATA_TYPE, "RaidificatorAchievementRow", 30, function(...) self:SetupItemRow(...) end)
	ZO_ScrollList_EnableHighlight(self.list, "ZO_ThinListHighlight")
	self:SetAlternateRowBackgrounds(true)

	local sortKeys = {
		["release"] = { isNumeric = true },
		["name"]    = { tiebreaker = "release", tieBreakerSortOrder = ZO_SORT_ORDER_UP },
		["clear"]   = { isNumeric = true, tiebreaker = "release", tieBreakerSortOrder = ZO_SORT_ORDER_UP },
		["hard"]    = { isNumeric = true, tiebreaker = "release", tieBreakerSortOrder = ZO_SORT_ORDER_UP },
		["speed"]   = { isNumeric = true, tiebreaker = "release", tieBreakerSortOrder = ZO_SORT_ORDER_UP },
		["vit"]     = { isNumeric = true, tiebreaker = "release", tieBreakerSortOrder = ZO_SORT_ORDER_UP },
		["tri"]     = { isNumeric = true, tiebreaker = "release", tieBreakerSortOrder = ZO_SORT_ORDER_UP },
		["count"]   = { isNumeric = true, tiebreaker = "release", tieBreakerSortOrder = ZO_SORT_ORDER_UP },
	}

	self.currentSortKey = "release"
	self.currentSortOrder = ZO_SORT_ORDER_DOWN
	self.sortHeaderGroup:SelectAndResetSortForKey(self.currentSortKey)
	self.sortFunction = function( listEntry1, listEntry2 )
		return ZO_TableOrderingFunction(listEntry1.data, listEntry2.data, self.currentSortKey, sortKeys, self.currentSortOrder)
	end

	self.filterDrop = ZO_ComboBox_ObjectFromContainer(self.frame:GetNamedChild("FilterDrop"))
	self:InitializeComboBox(self.filterDrop, { prefix = "SI_RCR_ACHIEVEMENT_CATEGORY", max = 3 }, RCR.vars.achCategoryId)

	self:InitializeDataSources()

	self:RefreshData()
end

function RaidificatorAchievementsList:BuildMasterList( )
	self.masterList = { }

	for idx, entry in ipairs(RCR.GetAchievementData()) do
		table.insert(self.masterList, {
			type = SORT_TYPE,
			release = idx,
			category = entry.category,
			year = entry.year,
			trophy = entry.trophy,
			name = entry.name,
			achs = entry.achs,
			contentId = entry.contentId,
		})
	end

	self:RefreshAchievementState(true)
end

function RaidificatorAchievementsList:RefreshAchievementState( initial )
	if (not self.masterList) then return end

	local stats = {
		seen = { },
		totals = { 0, 0, 0 },
		complete = { 0, 0, 0 },
	}

	for _, data in ipairs(self.masterList) do
		local totals = { A = 0, H = 0, S = 0, N = 0, T = 0 }
		local earned = { A = 0, H = 0, S = 0, N = 0, T = 0 }
		local clear = -1

		local ratio = function(k) return (totals[k] > 0) and (earned[k] / totals[k]) or -1 end

		for _, ach in ipairs(data.achs) do
			local completed = false

			totals.A = totals.A + 1
			AddToTally(totals, ach.flags)
			if (CheckFlag(ach.flags, "C")) then
				clear = zo_max(clear, 0)
			end

			if (IsAchComplete(ach.id)) then
				earned.A = earned.A + 1
				AddToTally(earned, ach.flags)
				if (CheckFlag(ach.flags, "C")) then
					clear = zo_max(clear, CheckFlag(ach.flags, "V") and DUNGEON_DIFFICULTY_VETERAN or DUNGEON_DIFFICULTY_NORMAL)
				end
				completed = true
			end

			if (not stats.seen[ach.id]) then
				stats.seen[ach.id] = true
				stats.totals[data.category] = stats.totals[data.category] + 1
				if (completed) then
					stats.complete[data.category] = stats.complete[data.category] + 1
				end
			end
		end

		data.clear = clear
		data.hard = ratio("H")
		data.speed = ratio("S")
		data.vit = ratio("N")
		data.tri = ratio("T")
		data.count = ratio("A")
		data.totals = totals
		data.earned = earned
	end

	self.stats = stats

	if (not initial) then
		self:RefreshFilters()
	end
end

function RaidificatorAchievementsList:FilterScrollList( )
	local scrollData = ZO_ScrollList_GetDataList(self.list)
	ZO_ClearNumericallyIndexedTable(scrollData)

	local filterId = self.filterDrop:GetSelectedItemData().id
	RCR.vars.achCategoryId = filterId

	for _, data in ipairs(self.masterList or { }) do
		if (filterId == data.category) then
			table.insert(scrollData, ZO_ScrollList_CreateDataEntry(DATA_TYPE, data))
		end
	end

	if (self.stats) then
		local total = self.stats.totals[filterId]
		local completed = self.stats.complete[filterId]
		self.frame:GetNamedChild("CompletedCount"):SetText(string.format(GetString(SI_RCR_COMPLETED_COUNT), completed, total, 100 * completed / total))
	end
end

function RaidificatorAchievementsList:SetupItemRow( control, data )
	local cell

	cell = control:GetNamedChild("Release")
	cell.normalColor = ZO_DEFAULT_TEXT
	cell:SetText(data.year)

	cell = control:GetNamedChild("Trophy")
	cell:SetTexture(GetCollectibleIcon(data.trophy))

	cell = control:GetNamedChild("Name")
	cell.normalColor = ZO_DEFAULT_TEXT
	cell:SetText(data.name)

	cell = control:GetNamedChild("Clear")
	if (data.clear == -1) then
		cell:SetText("")
	elseif (data.clear == 0) then
		cell:SetText(GetString(SI_ALLIANCE0))
		SetRatioColor(cell, data.clear)
	else
		cell:SetText(GetString("SI_DUNGEONDIFFICULTY", data.clear))
		SetRatioColor(cell, data.clear / 2)
	end

	cell = control:GetNamedChild("HardText")
	if (data.totals.H > 1) then
		cell:SetText(string.format("%d / %d", data.earned.H, data.totals.H))
		SetRatioColor(cell, data.hard)
	else
		cell:SetText("")
	end

	cell = control:GetNamedChild("HardIcon")
	cell:SetTexture(RCR.GetTexture("ach-hm"))
	SetRatioColor(cell, data.hard)

	cell = control:GetNamedChild("SpeedIcon")
	cell:SetTexture(RCR.GetTexture("ach-sr"))
	SetRatioColor(cell, data.speed)

	cell = control:GetNamedChild("VitIcon")
	cell:SetTexture(RCR.GetTexture("ach-nd"))
	SetRatioColor(cell, data.vit)

	cell = control:GetNamedChild("TriIcon")
	cell:SetTexture(RCR.GetTexture("ach-tri"))
	SetRatioColor(cell, data.tri)

	cell = control:GetNamedChild("Count")
	if (data.totals.A > 1) then
		cell:SetText(string.format("%d / %d", data.earned.A, data.totals.A))
		SetRatioColor(cell, data.count)
	else
		cell:SetText("")
	end

	self:SetupRow(control, data)
end

function RaidificatorAchievementsList:InitializeDataSources( )
	local server = LCCC.GetServerName()
	local userId = GetDisplayName()
	local accountLists = { [server] = { userId } }
	local servers = { server }

	self.server = server
	self.account = userId
	self.accountLists = accountLists

	if (LMAA) then
		for _, account in ipairs(LMAA.GetAccounts()) do
			if (account[1] ~= server or account[2] ~= userId) then
				if (not accountLists[account[1]]) then
					accountLists[account[1]] = { }
					table.insert(servers, account[1])
				end
				table.insert(accountLists[account[1]], account[2])
			end
		end

		GetAchTimestamp = function( achievementId )
			if (GetAchievementPersistenceLevel(achievementId) == ACHIEVEMENT_PERSISTENCE_CHARACTER) then
				local earliestTimestamp
				for _, character in ipairs(LMAA.GetCharacters(self.server, self.account)) do
					local timestamp = LMAA.GetAchievementTimestamp(character[1], achievementId)
					if (timestamp > 0 and (not earliestTimestamp or earliestTimestamp > timestamp)) then
						earliestTimestamp = timestamp
					end
				end
				return earliestTimestamp or 0
			else
				return LMAA.GetAchievementTimestamp({ self.server, self.account }, achievementId)
			end
		end

		IsAchComplete = function( achievementId )
			if (GetAchievementPersistenceLevel(achievementId) == ACHIEVEMENT_PERSISTENCE_CHARACTER) then
				for _, character in ipairs(LMAA.GetCharacters(self.server, self.account)) do
					if (LMAA.IsAchievementComplete(character[1], achievementId)) then
						return true
					end
				end
				return false
			else
				return LMAA.IsAchievementComplete({ self.server, self.account }, achievementId)
			end
		end

		GetAchCriterion = function(...) return LMAA.GetAchievementCriterion({ self.server, self.account }, ...) end
		GetAchLink = function(...) return LMAA.GetAchievementLink({ self.server, self.account }, ...) end
	end

	local control = self.frame:GetNamedChild("AccountDrop")
	control:GetNamedChild("Caption"):SetText(GetString(SI_LEJ_ACCOUNT))
	self.accountDrop = ZO_ComboBox_ObjectFromContainer(control)

	local control = self.frame:GetNamedChild("ServerDrop")
	control:GetNamedChild("Caption"):SetText(GetString(SI_LEJ_SERVER))
	self.serverDrop = ZO_ComboBox_ObjectFromContainer(control)
	self:InitializeComboBox(self.serverDrop, { list = servers }, nil, true, function( comboBox, entryText, entry, selectionChanged )
		self.server = entryText
		self:RefreshAccountList()
	end)
end

function RaidificatorAchievementsList:RefreshAccountList( )
	local accounts = self.accountLists[self.server]

	-- Try to keep the same account selected when changing servers
	local initialIndex
	for i, account in ipairs(accounts) do
		if (self.account == account) then
			initialIndex = i
		end
	end

	self:InitializeComboBox(self.accountDrop, { list = accounts }, initialIndex, true, function( comboBox, entryText, entry, selectionChanged )
		self.account = entryText
		self:RefreshAchievementState()
	end)
end


--------------------------------------------------------------------------------
-- Achievement Info Popup
--------------------------------------------------------------------------------

local RaidificatorAchievementInfo = ZO_Object:Subclass()

function RaidificatorAchievementInfo:New( )
	local obj = ZO_Object.New(self)
	obj.control = RaidificatorAchievementInfoTopLevel
	obj.PADDING = 12
	obj.R_WIDTH = 640
	obj.R_HEIGHT = 20
	obj.rows = { }
	obj.index = 1
	return obj
end

function RaidificatorAchievementInfo:GetNewRow( )
	local rows = self.rows
	local index = self.index
	self.index = index + 1

	-- Create new rows lazily, only as needed
	while (index > #rows) do
		local row = WINDOW_MANAGER:CreateControlFromVirtual("$(parent)Row" .. tostring(#rows + 1), self.control, "RaidificatorAchievementInfoRow")
		if (#rows == 0) then
			row:SetAnchor(TOPLEFT, nil, nil, self.PADDING, self.PADDING)
		else
			row:SetAnchor(TOPLEFT, rows[#rows], BOTTOMLEFT)
		end
		row:GetNamedChild("BG"):SetHidden(#rows % 2 == 0)
		row.elements = {
			icon = row:GetNamedChild("Icon"),
			name = row:GetNamedChild("Name"),
			progress = row:GetNamedChild("Progress"),
			rewards = row:GetNamedChild("Rewards"),
		}
		table.insert(rows, row)
	end

	rows[index]:SetHidden(false)
	return rows[index].elements
end

function RaidificatorAchievementInfo:Show( )
	local control = self.control
	local rows = self.rows
	local index = self.index

	-- Hide unused rows
	for i = index, #rows do
		rows[i]:SetHidden(true)
	end

	control:SetDimensions(self.R_WIDTH + self.PADDING * 2 + 4, self.R_HEIGHT * (index - 1) + self.PADDING * 2)
	control:ClearAnchors()
	control:SetAnchor(TOPRIGHT, LEJ.GetFrame(), TOPLEFT, -100, 0)
	control:SetHidden(false)
end

function RaidificatorAchievementInfo:Reset( )
	self.control:SetHidden(true)
	self.index = 1
end


--------------------------------------------------------------------------------
-- XML Handlers
--------------------------------------------------------------------------------

local Popup = RaidificatorAchievementInfo:New()

function RaidificatorAchievementRow_OnMouseEnter( control )
	local data = ZO_ScrollList_GetData(control)
	RCR.achList:Row_OnMouseEnter(control)

	local results = { }

	for _, ach in ipairs(data.achs) do
		local icon
		if (CheckFlag(ach.flags, "H")) then
			icon = RCR.GetTexture("ach-hm")
		elseif (CheckFlag(ach.flags, "S")) then
			icon = RCR.GetTexture("ach-sr")
		elseif (CheckFlag(ach.flags, "N")) then
			icon = RCR.GetTexture("ach-nd")
		elseif (CheckFlag(ach.flags, "T")) then
			icon = RCR.GetTexture("ach-tri")
		elseif (CheckFlag(ach.flags, "V")) then
			icon = "/esoui/art/lfg/gamepad/gp_lfg_menuicon_veterandungeon.dds"
		else
			icon = RCR.GetTexture()
		end

		local rewards = { }
		if (GetAchievementRewardTitle(ach.id)) then
			table.insert(rewards, zo_strformat("<<1>>|cFFFFFF<<2>>|r", GetString(SI_ACHIEVEMENTS_TITLE), select(2, GetAchievementRewardTitle(ach.id))))
		end
		if (GetAchievementRewardCollectible(ach.id)) then
			local name, _, _, _, _, _, _, categoryType = GetCollectibleInfo(select(2, GetAchievementRewardCollectible(ach.id)))
			table.insert(rewards, zo_strformat("<<1>>: |cFFFFFF<<t:2>>|r", GetString("SI_COLLECTIBLECATEGORYTYPE", categoryType), name))
		end
		if (GetAchievementRewardDye(ach.id)) then
			local name, _, _, _, _, r, g, b = GetDyeInfoById(select(2, GetAchievementRewardDye(ach.id)))
			table.insert(rewards, string.format("%s|c%06X%s|r", GetString(SI_ACHIEVEMENTS_DYE), LCCC.RGBToInt24(r, g, b), name))
		end

		table.insert(results, {
			id = ach.id,
			name = zo_strformat("<<1>>", GetAchievementName(ach.id)),
			time = GetAchTimestamp(ach.id),
			icon = icon,
			rewards = rewards,
		})
	end

	table.sort(results, function( a, b )
		if (a.time == b.time) then
			return a.name < b.name
		else
			return b.time < a.time
		end
	end)

	Popup:Reset()

	local lines = { }
	for i, result in ipairs(results) do
		local row = Popup:GetNewRow()
		row.icon:SetTexture(result.icon)
		row.name:SetText(result.name)
		if (result.time > 0) then
			row.name:SetColor(LCCC.Int32ToRGBA(COLOR_COMPLETE_POPUP))
			row.progress:SetText(RCR.FormatTimestamp(result.time))
			row.progress:SetColor(ZO_TOOLTIP_DEFAULT_COLOR:UnpackRGB())
		else
			local completed = 0
			local totalRequired = 0
			for i = 1, GetAchievementNumCriteria(result.id) do
				local _, numCompleted, numRequired = GetAchCriterion(result.id, i)
				completed = completed + numCompleted
				totalRequired = totalRequired + numRequired
			end

			row.name:SetColor(LCCC.Int32ToRGBA(COLOR_INCOMPLETE))
			row.progress:SetText((totalRequired > 1) and string.format("%d / %d", completed, totalRequired) or "—")
			row.progress:SetColor(LCCC.Int32ToRGBA(COLOR_INCOMPLETE))
		end
		row.rewards:SetText(table.concat(result.rewards, ", "))
	end

	Popup:Show()
end

function RaidificatorAchievementRow_OnMouseExit( control )
	RCR.achList:Row_OnMouseExit(control)

	Popup:Reset()
end

function RaidificatorAchievementRow_OnMouseUp( ... )
	RCR.achList:Row_OnMouseUp(...)
end
